@@ -1,12 +1,12 @@
 
 /* Compiler implementation of the D programming language
- * Copyright (c) 1999-2014 by Digital Mars
+ * Copyright (c) 1999-2016 by Digital Mars
  * All Rights Reserved
  * written by Walter Bright
  * http://www.digitalmars.com
  * Distributed under the Boost Software License, Version 1.0.
  * http://www.boost.org/LICENSE_1_0.txt
- * https://github.com/D-Programming-Language/dmd/blob/master/src/expression.h
+ * https://github.com/dlang/dmd/blob/master/src/expression.h
  */
 
 #ifndef DMD_EXPRESSION_H
@@ -79,7 +79,7 @@ Expression *doCopyOrMove(Scope *sc, Expression *e);
 Expression *resolveOpDollar(Scope *sc, ArrayExp *ae, Expression **pe0);
 Expression *resolveOpDollar(Scope *sc, ArrayExp *ae, IntervalExp *ie, Expression **pe0);
 Expression *integralPromotions(Expression *e, Scope *sc);
-void discardValue(Expression *e);
+bool discardValue(Expression *e);
 bool isTrivialExp(Expression *e);
 
 int isConst(Expression *e);
@@ -135,11 +135,6 @@ public:
     unsigned char size;         // # of bytes in Expression so we can copy() it
     unsigned char parens;       // if this is a parenthesized expression
 
-#if IN_LLVM
-    llvm::Value* cachedLvalue;
-#endif
-
-    Expression(Loc loc, TOK op, int size);
     static void _init();
     Expression *copy();
     virtual Expression *syntaxCopy();
@@ -150,7 +145,7 @@ public:
     int dyncast() { return DYNCAST_EXPRESSION; }
 
     void print();
-    char *toChars();
+    const char *toChars();
     virtual void printAST(int ident = 0);
     void error(const char *format, ...) const;
     void warning(const char *format, ...) const;
@@ -232,8 +227,6 @@ class IntegerExp : public Expression
 public:
     dinteger_t value;
 
-    IntegerExp(Loc loc, dinteger_t value, Type *type);
-    IntegerExp(dinteger_t value);
     bool equals(RootObject *o);
     Expression *semantic(Scope *sc);
     dinteger_t toInteger();
@@ -253,8 +246,6 @@ private:
 class ErrorExp : public Expression
 {
 public:
-    ErrorExp();
-
     Expression *toLvalue(Scope *sc, Expression *e);
     void accept(Visitor *v) { v->visit(this); }
 
@@ -266,7 +257,6 @@ class RealExp : public Expression
 public:
     real_t value;
 
-    RealExp(Loc loc, real_t value, Type *type);
     bool equals(RootObject *o);
     Expression *semantic(Scope *sc);
     dinteger_t toInteger();
@@ -283,7 +273,6 @@ class ComplexExp : public Expression
 public:
     complex_t value;
 
-    ComplexExp(Loc loc, complex_t value, Type *type);
     bool equals(RootObject *o);
     Expression *semantic(Scope *sc);
     dinteger_t toInteger();
@@ -299,9 +288,7 @@ class IdentifierExp : public Expression
 {
 public:
     Identifier *ident;
-    Declaration *var;
 
-    IdentifierExp(Loc loc, Identifier *ident);
     static IdentifierExp *create(Loc loc, Identifier *ident);
     Expression *semantic(Scope *sc);
     bool isLvalue();
@@ -312,7 +299,6 @@ public:
 class DollarExp : public IdentifierExp
 {
 public:
-    DollarExp(Loc loc);
     void accept(Visitor *v) { v->visit(this); }
 };
 
@@ -322,9 +308,8 @@ public:
     Dsymbol *s;
     bool hasOverloads;
 
-    DsymbolExp(Loc loc, Dsymbol *s, bool hasOverloads = true);
     Expression *semantic(Scope *sc);
-    static Expression resolve(Loc loc, Scope *sc, Dsymbol *s, bool hasOverloads);
+    static Expression *resolve(Loc loc, Scope *sc, Dsymbol *s, bool hasOverloads);
     bool isLvalue();
     Expression *toLvalue(Scope *sc, Expression *e);
     void accept(Visitor *v) { v->visit(this); }
@@ -335,7 +320,6 @@ class ThisExp : public Expression
 public:
     VarDeclaration *var;
 
-    ThisExp(Loc loc);
     Expression *semantic(Scope *sc);
     bool isBool(bool result);
     bool isLvalue();
@@ -347,7 +331,6 @@ public:
 class SuperExp : public ThisExp
 {
 public:
-    SuperExp(Loc loc);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -358,7 +341,6 @@ class NullExp : public Expression
 public:
     unsigned char committed;    // !=0 if type is committed
 
-    NullExp(Loc loc, Type *t = NULL);
     bool equals(RootObject *o);
     Expression *semantic(Scope *sc);
     bool isBool(bool result);
@@ -368,17 +350,14 @@ public:
 
 class StringExp : public Expression
 {
+public:
     void *string;       // char, wchar, or dchar data
     size_t len;         // number of chars, wchars, or dchars
-public:
     unsigned char sz;   // 1: char, 2: wchar, 4: dchar
     unsigned char committed;    // !=0 if type is committed
     utf8_t postfix;      // 'c', 'w', 'd'
     OwnedBy ownedByCtfe;
 
-    StringExp(Loc loc, char *s);
-    StringExp(Loc loc, void *s, size_t len);
-    StringExp(Loc loc, void *s, size_t len, utf8_t postfix);
     static StringExp *create(Loc loc, char *s);
     bool equals(RootObject *o);
     Expression *semantic(Scope *sc);
@@ -391,7 +370,16 @@ public:
     Expression *modifiableLvalue(Scope *sc, Expression *e);
     unsigned charAt(uinteger_t i) const;
     void accept(Visitor *v) { v->visit(this); }
-    const char *toStringz() const;
+#if IN_LLVM
+    // The D version returns a slice.
+    const char *toStringz() const
+    {
+        auto nbytes = len * sz;
+        char* s = (char*)mem.xmalloc(nbytes + sz);
+        writeTo(s, true);
+        return s;
+    }
+#endif
     size_t numberOfCodeUnits(int tynto = 0) const;
     void writeTo(void* dest, bool zero, int tyto = 0) const;
     char *toPtr();
@@ -412,9 +400,6 @@ public:
      */
     Expressions *exps;
 
-    TupleExp(Loc loc, Expression *e0, Expressions *exps);
-    TupleExp(Loc loc, Expressions *exps);
-    TupleExp(Loc loc, TupleDeclaration *tup);
     Expression *syntaxCopy();
     bool equals(RootObject *o);
     Expression *semantic(Scope *sc);
@@ -429,16 +414,10 @@ public:
     Expressions *elements;
     OwnedBy ownedByCtfe;
 
-    ArrayLiteralExp(Loc loc, Expressions *elements);
-    ArrayLiteralExp(Loc loc, Expression *e);
-    ArrayLiteralExp(Loc loc, Expression *basis, Expressions *elements);
-    ArrayLiteralExp(Loc loc, Expression *basis, size_t dim);
-
     Expression *syntaxCopy();
     bool equals(RootObject *o);
     Expression *getElement(d_size_t i);
     static Expressions* copyElements(Expression *e1, Expression *e2 = NULL);
-    Expressions *copyElements();
     Expression *semantic(Scope *sc);
     bool isBool(bool result);
     StringExp *toStringExp();
@@ -453,7 +432,6 @@ public:
     Expressions *values;
     OwnedBy ownedByCtfe;
 
-    AssocArrayLiteralExp(Loc loc, Expressions *keys, Expressions *values);
     bool equals(RootObject *o);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
@@ -495,7 +473,7 @@ public:
 
     bool useStaticInit;         // if this is true, use the StructDeclaration's init symbol
     Symbol *sym;                // back end symbol to initialize with literal
-    bool fillHoles;             // fill alignment 'holes' with zero
+
     OwnedBy ownedByCtfe;
 
     // pointer to the origin instance of the expression.
@@ -513,7 +491,6 @@ public:
     // (with infinite recursion) of this expression.
     int stageflags;
 
-    StructLiteralExp(Loc loc, StructDeclaration *sd, Expressions *elements, Type *stype = NULL);
     static StructLiteralExp *create(Loc loc, StructDeclaration *sd, void *elements, Type *stype = NULL);
     bool equals(RootObject *o);
     Expression *syntaxCopy();
@@ -531,7 +508,6 @@ DotIdExp *typeDotIdExp(Loc loc, Type *type, Identifier *ident);
 class TypeExp : public Expression
 {
 public:
-    TypeExp(Loc loc, Type *type);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     bool checkType();
@@ -544,9 +520,10 @@ class ScopeExp : public Expression
 public:
     ScopeDsymbol *sds;
 
-    ScopeExp(Loc loc, ScopeDsymbol *sds);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
+    bool checkType();
+    bool checkValue();
     void accept(Visitor *v) { v->visit(this); }
 };
 
@@ -556,7 +533,6 @@ public:
     TemplateDeclaration *td;
     FuncDeclaration *fd;
 
-    TemplateExp(Loc loc, TemplateDeclaration *td, FuncDeclaration *fd = NULL);
     bool isLvalue();
     Expression *toLvalue(Scope *sc, Expression *e);
     bool checkType();
@@ -580,8 +556,6 @@ public:
     NewDeclaration *allocator;  // allocator function
     int onstack;                // allocate on stack
 
-    NewExp(Loc loc, Expression *thisexp, Expressions *newargs,
-        Type *newtype, Expressions *arguments);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
 
@@ -598,8 +572,6 @@ public:
     ClassDeclaration *cd;       // class being instantiated
     Expressions *arguments;     // Array of Expression's to call class constructor
 
-    NewAnonClassExp(Loc loc, Expression *thisexp, Expressions *newargs,
-        ClassDeclaration *cd, Expressions *arguments);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -611,8 +583,6 @@ public:
     Declaration *var;
     bool hasOverloads;
 
-    SymbolExp(Loc loc, TOK op, int size, Declaration *var, bool hasOverloads);
-
     void accept(Visitor *v) { v->visit(this); }
 };
 
@@ -623,7 +593,6 @@ class SymOffExp : public SymbolExp
 public:
     dinteger_t offset;
 
-    SymOffExp(Loc loc, Declaration *var, dinteger_t offset, bool hasOverloads = true);
     Expression *semantic(Scope *sc);
     bool isBool(bool result);
 
@@ -635,7 +604,6 @@ public:
 class VarExp : public SymbolExp
 {
 public:
-    VarExp(Loc loc, Declaration *var, bool hasOverloads = true);
     static VarExp *create(Loc loc, Declaration *var, bool hasOverloads = true);
     bool equals(RootObject *o);
     Expression *semantic(Scope *sc);
@@ -655,7 +623,6 @@ class OverExp : public Expression
 public:
     OverloadSet *vars;
 
-    OverExp(Loc loc, OverloadSet *s);
     bool isLvalue();
     Expression *toLvalue(Scope *sc, Expression *e);
     void accept(Visitor *v) { v->visit(this); }
@@ -670,13 +637,12 @@ public:
     TemplateDeclaration *td;
     TOK tok;
 
-    FuncExp(Loc loc, Dsymbol *s);
     void genIdent(Scope *sc);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     Expression *semantic(Scope *sc, Expressions *arguments);
     MATCH matchType(Type *to, Scope *sc, FuncExp **pfe, int flag = 0);
-    char *toChars();
+    const char *toChars();
     bool checkType();
     bool checkValue();
 
@@ -693,7 +659,6 @@ class DeclarationExp : public Expression
 public:
     Dsymbol *declaration;
 
-    DeclarationExp(Loc loc, Dsymbol *declaration);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
 
@@ -705,7 +670,6 @@ class TypeidExp : public Expression
 public:
     RootObject *obj;
 
-    TypeidExp(Loc loc, RootObject *obj);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -717,7 +681,6 @@ public:
     Identifier *ident;
     Objects *args;
 
-    TraitsExp(Loc loc, Identifier *ident, Objects *args);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -726,7 +689,6 @@ public:
 class HaltExp : public Expression
 {
 public:
-    HaltExp(Loc loc);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -742,11 +704,9 @@ public:
     Identifier *id;     // can be NULL
     TOK tok;       // ':' or '=='
     Type *tspec;        // can be NULL
-    TOK tok2;      // 'struct', 'union', 'typedef', etc.
+    TOK tok2;      // 'struct', 'union', etc.
     TemplateParameters *parameters;
 
-    IsExp(Loc loc, Type *targ, Identifier *id, TOK tok, Type *tspec,
-        TOK tok2, TemplateParameters *parameters);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -760,7 +720,6 @@ public:
     Expression *e1;
     Type *att1; // Save alias this type to detect recursion
 
-    UnaExp(Loc loc, TOK op, int size, Expression *e1);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc) = 0;
     Expression *unaSemantic(Scope *sc);
@@ -782,7 +741,6 @@ public:
     Type *att1; // Save alias this type to detect recursion
     Type *att2; // Save alias this type to detect recursion
 
-    BinExp(Loc loc, TOK op, int size, Expression *e1, Expression *e2);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc) = 0;
     Expression *binSemantic(Scope *sc);
@@ -801,11 +759,6 @@ public:
 class BinAssignExp : public BinExp
 {
 public:
-    BinAssignExp(Loc loc, TOK op, int size, Expression *e1, Expression *e2)
-        : BinExp(loc, op, size, e1, e2)
-    {
-    }
-
     Expression *semantic(Scope *sc);
 
     bool isLvalue();
@@ -819,7 +772,6 @@ public:
 class CompileExp : public UnaExp
 {
 public:
-    CompileExp(Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -827,7 +779,6 @@ public:
 class ImportExp : public UnaExp
 {
 public:
-    ImportExp(Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -837,7 +788,6 @@ class AssertExp : public UnaExp
 public:
     Expression *msg;
 
-    AssertExp(Loc loc, Expression *e, Expression *msg = NULL);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
 
@@ -849,7 +799,6 @@ class DotIdExp : public UnaExp
 public:
     Identifier *ident;
 
-    DotIdExp(Loc loc, Expression *e, Identifier *ident);
     static DotIdExp *create(Loc loc, Expression *e, Identifier *ident);
     Expression *semantic(Scope *sc);
     Expression *semanticX(Scope *sc);
@@ -862,7 +811,6 @@ class DotTemplateExp : public UnaExp
 public:
     TemplateDeclaration *td;
 
-    DotTemplateExp(Loc loc, Expression *e, TemplateDeclaration *td);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -873,7 +821,6 @@ public:
     Declaration *var;
     bool hasOverloads;
 
-    DotVarExp(Loc loc, Expression *e, Declaration *var, bool hasOverloads = true);
     Expression *semantic(Scope *sc);
     int checkModifiable(Scope *sc, int flag);
     bool checkReadModifyWrite();
@@ -888,8 +835,6 @@ class DotTemplateInstanceExp : public UnaExp
 public:
     TemplateInstance *ti;
 
-    DotTemplateInstanceExp(Loc loc, Expression *e, Identifier *name, Objects *tiargs);
-    DotTemplateInstanceExp(Loc loc, Expression *e, TemplateInstance *ti);
     Expression *syntaxCopy();
     bool findTempDecl(Scope *sc);
     Expression *semantic(Scope *sc);
@@ -903,7 +848,6 @@ public:
     FuncDeclaration *func;
     bool hasOverloads;
 
-    DelegateExp(Loc loc, Expression *e, FuncDeclaration *func, bool hasOverloads = true);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -914,7 +858,6 @@ class DotTypeExp : public UnaExp
 public:
     Dsymbol *sym;               // symbol that represents a type
 
-    DotTypeExp(Loc loc, Expression *e, Dsymbol *sym);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -926,11 +869,6 @@ public:
     FuncDeclaration *f;         // symbol to call
     bool directcall;            // true if a virtual call is devirtualized
 
-    CallExp(Loc loc, Expression *e, Expressions *exps);
-    CallExp(Loc loc, Expression *e);
-    CallExp(Loc loc, Expression *e, Expression *earg1);
-    CallExp(Loc loc, Expression *e, Expression *earg1, Expression *earg2);
-
     static CallExp *create(Loc loc, Expression *e, Expressions *exps);
     static CallExp *create(Loc loc, Expression *e);
     static CallExp *create(Loc loc, Expression *e, Expression *earg1);
@@ -947,7 +885,6 @@ public:
 class AddrExp : public UnaExp
 {
 public:
-    AddrExp(Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -956,8 +893,6 @@ public:
 class PtrExp : public UnaExp
 {
 public:
-    PtrExp(Loc loc, Expression *e);
-    PtrExp(Loc loc, Expression *e, Type *t);
     Expression *semantic(Scope *sc);
     int checkModifiable(Scope *sc, int flag);
     bool isLvalue();
@@ -970,7 +905,6 @@ public:
 class NegExp : public UnaExp
 {
 public:
-    NegExp(Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -979,7 +913,6 @@ public:
 class UAddExp : public UnaExp
 {
 public:
-    UAddExp(Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -988,7 +921,6 @@ public:
 class ComExp : public UnaExp
 {
 public:
-    ComExp(Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -997,23 +929,14 @@ public:
 class NotExp : public UnaExp
 {
 public:
-    NotExp(Loc loc, Expression *e);
-    Expression *semantic(Scope *sc);
-    void accept(Visitor *v) { v->visit(this); }
-};
-
-class BoolExp : public UnaExp
-{
-public:
-    BoolExp(Loc loc, Expression *e, Type *type);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class DeleteExp : public UnaExp
 {
+    bool isRAII;
 public:
-    DeleteExp(Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
     Expression *toBoolean(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1026,8 +949,6 @@ public:
     Type *to;                   // type to cast to
     unsigned char mod;          // MODxxxxx
 
-    CastExp(Loc loc, Expression *e, Type *t);
-    CastExp(Loc loc, Expression *e, unsigned char mod);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
 
@@ -1040,7 +961,6 @@ public:
     TypeVector *to;             // the target vector type before semantic()
     unsigned dim;               // number of elements in the vector
 
-    VectorExp(Loc loc, Expression *e, Type *t);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1055,8 +975,6 @@ public:
     bool upperIsInBounds;       // true if upr <= e1.length
     bool lowerIsLessThanUpper;  // true if lwr <= upr
 
-    SliceExp(Loc loc, Expression *e1, IntervalExp *ie);
-    SliceExp(Loc loc, Expression *e1, Expression *lwr, Expression *upr);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     int checkModifiable(Scope *sc, int flag);
@@ -1071,7 +989,6 @@ public:
 class ArrayLengthExp : public UnaExp
 {
 public:
-    ArrayLengthExp(Loc loc, Expression *e1);
     Expression *semantic(Scope *sc);
 
     static Expression *rewriteOpAssign(BinExp *exp);
@@ -1084,7 +1001,6 @@ public:
     Expression *lwr;
     Expression *upr;
 
-    IntervalExp(Loc loc, Expression *lwr, Expression *upr);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1093,7 +1009,6 @@ public:
 class DelegatePtrExp : public UnaExp
 {
 public:
-    DelegatePtrExp(Loc loc, Expression *e1);
     Expression *semantic(Scope *sc);
     bool isLvalue();
     Expression *toLvalue(Scope *sc, Expression *e);
@@ -1103,7 +1018,6 @@ public:
 class DelegateFuncptrExp : public UnaExp
 {
 public:
-    DelegateFuncptrExp(Loc loc, Expression *e1);
     Expression *semantic(Scope *sc);
     bool isLvalue();
     Expression *toLvalue(Scope *sc, Expression *e);
@@ -1119,8 +1033,6 @@ public:
     size_t currentDimension;            // for opDollar
     VarDeclaration *lengthVar;
 
-    ArrayExp(Loc loc, Expression *e1, Expression *index = NULL);
-    ArrayExp(Loc loc, Expression *e1, Expressions *args);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     bool isLvalue();
@@ -1134,7 +1046,6 @@ public:
 class DotExp : public BinExp
 {
 public:
-    DotExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -1142,7 +1053,8 @@ public:
 class CommaExp : public BinExp
 {
 public:
-    CommaExp(Loc loc, Expression *e1, Expression *e2);
+    bool isGenerated;
+    bool allowCommaExp;
     Expression *semantic(Scope *sc);
     int checkModifiable(Scope *sc, int flag);
     bool isLvalue();
@@ -1160,7 +1072,6 @@ public:
     bool modifiable;
     bool indexIsInBounds;       // true if 0 <= e2 && e2 <= e1.length - 1
 
-    IndexExp(Loc loc, Expression *e1, Expression *e2);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     int checkModifiable(Scope *sc, int flag);
@@ -1178,7 +1089,6 @@ public:
 class PostExp : public BinExp
 {
 public:
-    PostExp(TOK op, Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -1188,7 +1098,6 @@ public:
 class PreExp : public UnaExp
 {
 public:
-    PreExp(TOK op, Loc loc, Expression *e);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -1204,7 +1113,6 @@ class AssignExp : public BinExp
 public:
     int memset;         // combination of MemorySet flags
 
-    AssignExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
     bool isLvalue();
     Expression *toLvalue(Scope *sc, Expression *ex);
@@ -1216,79 +1124,66 @@ public:
 class ConstructExp : public AssignExp
 {
 public:
-    ConstructExp(Loc loc, Expression *e1, Expression *e2);
-    ConstructExp(Loc loc, VarDeclaration *v, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class BlitExp : public AssignExp
 {
 public:
-    BlitExp(Loc loc, Expression *e1, Expression *e2);
-    BlitExp(Loc loc, VarDeclaration *v, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class AddAssignExp : public BinAssignExp
 {
 public:
-    AddAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class MinAssignExp : public BinAssignExp
 {
 public:
-    MinAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class MulAssignExp : public BinAssignExp
 {
 public:
-    MulAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class DivAssignExp : public BinAssignExp
 {
 public:
-    DivAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class ModAssignExp : public BinAssignExp
 {
 public:
-    ModAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class AndAssignExp : public BinAssignExp
 {
 public:
-    AndAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class OrAssignExp : public BinAssignExp
 {
 public:
-    OrAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class XorAssignExp : public BinAssignExp
 {
 public:
-    XorAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class PowAssignExp : public BinAssignExp
 {
 public:
-    PowAssignExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -1296,28 +1191,24 @@ public:
 class ShlAssignExp : public BinAssignExp
 {
 public:
-    ShlAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class ShrAssignExp : public BinAssignExp
 {
 public:
-    ShrAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class UshrAssignExp : public BinAssignExp
 {
 public:
-    UshrAssignExp(Loc loc, Expression *e1, Expression *e2);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class CatAssignExp : public BinAssignExp
 {
 public:
-    CatAssignExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -1325,7 +1216,6 @@ public:
 class AddExp : public BinExp
 {
 public:
-    AddExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1334,7 +1224,6 @@ public:
 class MinExp : public BinExp
 {
 public:
-    MinExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1343,7 +1232,6 @@ public:
 class CatExp : public BinExp
 {
 public:
-    CatExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1352,7 +1240,6 @@ public:
 class MulExp : public BinExp
 {
 public:
-    MulExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1361,7 +1248,6 @@ public:
 class DivExp : public BinExp
 {
 public:
-    DivExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1370,7 +1256,6 @@ public:
 class ModExp : public BinExp
 {
 public:
-    ModExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1379,7 +1264,6 @@ public:
 class PowExp : public BinExp
 {
 public:
-    PowExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1388,7 +1272,6 @@ public:
 class ShlExp : public BinExp
 {
 public:
-    ShlExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1397,7 +1280,6 @@ public:
 class ShrExp : public BinExp
 {
 public:
-    ShrExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1406,7 +1288,6 @@ public:
 class UshrExp : public BinExp
 {
 public:
-    UshrExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1415,7 +1296,6 @@ public:
 class AndExp : public BinExp
 {
 public:
-    AndExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1424,7 +1304,6 @@ public:
 class OrExp : public BinExp
 {
 public:
-    OrExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1433,7 +1312,6 @@ public:
 class XorExp : public BinExp
 {
 public:
-    XorExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1442,7 +1320,6 @@ public:
 class OrOrExp : public BinExp
 {
 public:
-    OrOrExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
     Expression *toBoolean(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1451,7 +1328,6 @@ public:
 class AndAndExp : public BinExp
 {
 public:
-    AndAndExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
     Expression *toBoolean(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1460,7 +1336,6 @@ public:
 class CmpExp : public BinExp
 {
 public:
-    CmpExp(TOK op, Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1469,7 +1344,6 @@ public:
 class InExp : public BinExp
 {
 public:
-    InExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1478,7 +1352,6 @@ public:
 class RemoveExp : public BinExp
 {
 public:
-    RemoveExp(Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -1488,7 +1361,6 @@ public:
 class EqualExp : public BinExp
 {
 public:
-    EqualExp(TOK op, Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
 
     void accept(Visitor *v) { v->visit(this); }
@@ -1499,7 +1371,6 @@ public:
 class IdentityExp : public BinExp
 {
 public:
-    IdentityExp(TOK op, Loc loc, Expression *e1, Expression *e2);
     Expression *semantic(Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
 };
@@ -1511,7 +1382,6 @@ class CondExp : public BinExp
 public:
     Expression *econd;
 
-    CondExp(Loc loc, Expression *econd, Expression *e1, Expression *e2);
     Expression *syntaxCopy();
     Expression *semantic(Scope *sc);
     int checkModifiable(Scope *sc, int flag);
@@ -1531,14 +1401,12 @@ class DefaultInitExp : public Expression
 public:
     TOK subop;             // which of the derived classes this is
 
-    DefaultInitExp(Loc loc, TOK subop, int size);
     void accept(Visitor *v) { v->visit(this); }
 };
 
 class FileInitExp : public DefaultInitExp
 {
 public:
-    FileInitExp(Loc loc);
     Expression *semantic(Scope *sc);
     Expression *resolveLoc(Loc loc, Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1547,7 +1415,6 @@ public:
 class LineInitExp : public DefaultInitExp
 {
 public:
-    LineInitExp(Loc loc);
     Expression *semantic(Scope *sc);
     Expression *resolveLoc(Loc loc, Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1556,7 +1423,6 @@ public:
 class ModuleInitExp : public DefaultInitExp
 {
 public:
-    ModuleInitExp(Loc loc);
     Expression *semantic(Scope *sc);
     Expression *resolveLoc(Loc loc, Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1565,7 +1431,6 @@ public:
 class FuncInitExp : public DefaultInitExp
 {
 public:
-    FuncInitExp(Loc loc);
     Expression *semantic(Scope *sc);
     Expression *resolveLoc(Loc loc, Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1574,7 +1439,6 @@ public:
 class PrettyFuncInitExp : public DefaultInitExp
 {
 public:
-    PrettyFuncInitExp(Loc loc);
     Expression *semantic(Scope *sc);
     Expression *resolveLoc(Loc loc, Scope *sc);
     void accept(Visitor *v) { v->visit(this); }
@@ -1623,7 +1487,7 @@ private:
         char sliceexp  [sizeof(SliceExp)];
 
         // Ensure that the union is suitably aligned.
-        longdouble for_alignment_only;
+        real_t for_alignment_only;
     } u;
 };
 
@@ -1672,4 +1536,5 @@ void sliceAssignStringFromString(StringExp *existingSE, StringExp *newstr, size_
 int sliceCmpStringWithString(StringExp *se1, StringExp *se2, size_t lo1, size_t lo2, size_t len);
 int sliceCmpStringWithArray(StringExp *se1, ArrayLiteralExp *ae2, size_t lo1, size_t lo2, size_t len);
 
+
 #endif /* DMD_EXPRESSION_H */
