@@ -1,12 +1,12 @@
 
 /* Compiler implementation of the D programming language
- * Copyright (c) 1999-2014 by Digital Mars
+ * Copyright (c) 1999-2016 by Digital Mars
  * All Rights Reserved
  * written by Walter Bright
  * http://www.digitalmars.com
  * Distributed under the Boost Software License, Version 1.0.
  * http://www.boost.org/LICENSE_1_0.txt
- * https://github.com/D-Programming-Language/dmd/blob/master/src/identifier.h
+ * https://github.com/dlang/dmd/blob/master/src/identifier.h
  */
 
 #ifndef DMD_IDENTIFIER_H
@@ -21,17 +21,18 @@
 
 class Identifier : public RootObject
 {
-public:
+private:
     int value;
     const char *string;
     size_t len;
 
-    Identifier(const char *string, int value);
-    static Identifier* create(const char *string, int value);
+public:
+    static Identifier* create(const char *string);
     bool equals(RootObject *o);
     int compare(RootObject *o);
     void print();
-    char *toChars();
+    const char *toChars();
+    int getValue() const;
     const char *toHChars2();
     int dyncast();
 
