@@ -23,11 +23,13 @@
 #include "llvm/Support/TargetSelect.h"
 #include "llvm/Target/TargetMachine.h"
 #include "llvm/Target/TargetOptions.h"
+#include "llvm/IR/Module.h"
 #include "mars.h"
 #include "gen/logger.h"
 
 #if LDC_LLVM_VER >= 307
 #include "driver/cl_options.h"
+#include "llvm/Support/TargetParser.h"
 
 static const char *getABI(const llvm::Triple &triple) {
   llvm::StringRef ABIName(opts::mABI);
@@ -88,24 +90,23 @@ MipsABI::Type getMipsABI() {
   // eabi can only be set on the commandline
   if (strncmp(opts::mABI.c_str(), "eabi", 4) == 0)
     return MipsABI::EABI;
-  else {
+
 #if LDC_LLVM_VER >= 308
-    const llvm::DataLayout dl = gTargetMachine->createDataLayout();
+  const llvm::DataLayout dl = gTargetMachine->createDataLayout();
 #else
-    const llvm::DataLayout &dl = *gTargetMachine->getDataLayout();
+  const llvm::DataLayout &dl = *gTargetMachine->getDataLayout();
 #endif
-    if (dl.getPointerSizeInBits() == 64)
-      return MipsABI::N64;
+
+  if (dl.getPointerSizeInBits() == 64)
+    return MipsABI::N64;
+
 #if LDC_LLVM_VER >= 309
-    else if (dl.getLargestLegalIntTypeSizeInBits() == 64)
+  const auto largestInt = dl.getLargestLegalIntTypeSizeInBits();
 #else
-    else if (dl.getLargestLegalIntTypeSize() == 64)
+  const auto largestInt = dl.getLargestLegalIntTypeSize();
 #endif
-      return MipsABI::N32;
-    else
-      return MipsABI::O32;
-  }
-#else
+  return (largestInt == 64) ? MipsABI::N32 : MipsABI::O32;
+#else // LDC_LLVM_VER < 307
   llvm::StringRef features = gTargetMachine->getTargetFeatureString();
   if (features.find("+o32") != std::string::npos) {
     return MipsABI::O32;
@@ -166,44 +167,56 @@ static std::string getX86TargetCPU(const llvm::Triple &triple) {
 }
 
 static std::string getARMTargetCPU(const llvm::Triple &triple) {
-  const char *result = llvm::StringSwitch<const char *>(triple.getArchName())
-                           .Cases("armv2", "armv2a", "arm2")
-                           .Case("armv3", "arm6")
-                           .Case("armv3m", "arm7m")
-                           .Case("armv4", "strongarm")
-                           .Case("armv4t", "arm7tdmi")
-                           .Cases("armv5", "armv5t", "arm10tdmi")
-                           .Cases("armv5e", "armv5te", "arm1026ejs")
-                           .Case("armv5tej", "arm926ej-s")
-                           .Cases("armv6", "armv6k", "arm1136jf-s")
-                           .Case("armv6j", "arm1136j-s")
-                           .Cases("armv6z", "armv6zk", "arm1176jzf-s")
-                           .Case("armv6t2", "arm1156t2-s")
-                           .Cases("armv6m", "armv6-m", "cortex-m0")
-                           .Cases("armv7", "armv7a", "armv7-a", "cortex-a8")
-                           .Cases("armv7l", "armv7-l", "cortex-a8")
-                           .Cases("armv7f", "armv7-f", "cortex-a9-mp")
-                           .Cases("armv7s", "armv7-s", "swift")
-                           .Cases("armv7r", "armv7-r", "cortex-r4")
-                           .Cases("armv7m", "armv7-m", "cortex-m3")
-                           .Cases("armv7em", "armv7e-m", "cortex-m4")
-                           .Cases("armv8", "armv8a", "armv8-a", "cortex-a53")
-                           .Case("ep9312", "ep9312")
-                           .Case("iwmmxt", "iwmmxt")
-                           .Case("xscale", "xscale")
-                           // If all else failed, return the most base CPU with
-                           // thumb interworking
-                           // supported by LLVM.
-                           .Default(nullptr);
-
-  if (result) {
-    return result;
-  }
+#if LDC_LLVM_VER >= 308
+  auto defaultCPU = llvm::ARM::getDefaultCPU(triple.getArchName());
+  if (!defaultCPU.empty())
+    return defaultCPU;
+#else
+  auto defaultCPU = llvm::StringSwitch<const char *>(triple.getArchName())
+                        .Cases("armv2", "armv2a", "arm2")
+                        .Case("armv3", "arm6")
+                        .Case("armv3m", "arm7m")
+                        .Case("armv4", "strongarm")
+                        .Case("armv4t", "arm7tdmi")
+                        .Cases("armv5", "armv5t", "arm10tdmi")
+                        .Cases("armv5e", "armv5te", "arm1026ejs")
+                        .Case("armv5tej", "arm926ej-s")
+                        .Cases("armv6", "armv6k", "arm1136jf-s")
+                        .Case("armv6j", "arm1136j-s")
+                        .Cases("armv6z", "armv6zk", "arm1176jzf-s")
+                        .Case("armv6t2", "arm1156t2-s")
+                        .Cases("armv6m", "armv6-m", "cortex-m0")
+                        .Cases("armv7", "armv7a", "armv7-a", "cortex-a8")
+                        .Cases("armv7l", "armv7-l", "cortex-a8")
+                        .Cases("armv7f", "armv7-f", "cortex-a9-mp")
+                        .Cases("armv7s", "armv7-s", "swift")
+                        .Cases("armv7r", "armv7-r", "cortex-r4")
+                        .Cases("armv7m", "armv7-m", "cortex-m3")
+                        .Cases("armv7em", "armv7e-m", "cortex-m4")
+                        .Cases("armv8", "armv8a", "armv8-a", "cortex-a53")
+                        .Case("ep9312", "ep9312")
+                        .Case("iwmmxt", "iwmmxt")
+                        .Case("xscale", "xscale")
+                        .Default(nullptr);
+  if (defaultCPU)
+    return defaultCPU;
+#endif
 
+  // Return the most base CPU with thumb interworking supported by LLVM.
   return (triple.getEnvironment() == llvm::Triple::GNUEABIHF) ? "arm1176jzf-s"
                                                               : "arm7tdmi";
 }
 
+static std::string getAArch64TargetCPU(const llvm::Triple &triple) {
+#if LDC_LLVM_VER >= 309
+  auto defaultCPU = llvm::AArch64::getDefaultCPU(triple.getArchName());
+  if (!defaultCPU.empty())
+    return defaultCPU;
+#endif
+
+  return "generic";
+}
+
 /// Returns the LLVM name of the target CPU to use given the provided
 /// -mcpu argument and target triple.
 static std::string getTargetCPU(const std::string &cpu,
@@ -230,7 +243,16 @@ static std::string getTargetCPU(const std::string &cpu,
   case llvm::Triple::x86_64:
     return getX86TargetCPU(triple);
   case llvm::Triple::arm:
+  case llvm::Triple::armeb:
+  case llvm::Triple::thumb:
     return getARMTargetCPU(triple);
+#if LDC_LLVM_VER == 305
+  case llvm::Triple::arm64:
+  case llvm::Triple::arm64_be:
+#endif
+  case llvm::Triple::aarch64:
+  case llvm::Triple::aarch64_be:
+    return getAArch64TargetCPU(triple);
   }
 }
 
@@ -424,17 +446,18 @@ const llvm::Target *lookupTarget(const std::string &arch, llvm::Triple &triple,
   return target;
 }
 
-llvm::TargetMachine *createTargetMachine(
-    std::string targetTriple, std::string arch, std::string cpu,
-    std::vector<std::string> attrs, ExplicitBitness::Type bitness,
-    FloatABI::Type floatABI,
+llvm::TargetMachine *
+createTargetMachine(std::string targetTriple, std::string arch, std::string cpu,
+                    std::vector<std::string> attrs,
+                    ExplicitBitness::Type bitness, FloatABI::Type floatABI,
 #if LDC_LLVM_VER >= 309
-    llvm::Optional<llvm::Reloc::Model> relocModel,
+                    llvm::Optional<llvm::Reloc::Model> relocModel,
 #else
-    llvm::Reloc::Model relocModel,
+                    llvm::Reloc::Model relocModel,
 #endif
-    llvm::CodeModel::Model codeModel, llvm::CodeGenOpt::Level codeGenOptLevel,
-    bool noFramePointerElim, bool noLinkerStripDead) {
+                    llvm::CodeModel::Model codeModel,
+                    llvm::CodeGenOpt::Level codeGenOptLevel,
+                    bool noFramePointerElim, bool noLinkerStripDead) {
   // Determine target triple. If the user didn't explicitly specify one, use
   // the one set at LLVM configure time.
   llvm::Triple triple;
@@ -518,7 +541,7 @@ llvm::TargetMachine *createTargetMachine(
                     features.getString().c_str());
   }
 
-  // Handle cases where LLVM picks wrong default relocModel
+// Handle cases where LLVM picks wrong default relocModel
 #if LDC_LLVM_VER >= 309
   if (!relocModel.hasValue()) {
 #else
@@ -528,7 +551,12 @@ llvm::TargetMachine *createTargetMachine(
       // Darwin defaults to PIC (and as of 10.7.5/LLVM 3.1-3.3, TLS use leads
       // to crashes for non-PIC code). LLVM doesn't handle this.
       relocModel = llvm::Reloc::PIC_;
-    } else if (triple.getEnvironment() == llvm::Triple::Android) {
+    } else if (triple.isOSLinux()) {
+      // Modern Linux distributions have their toolchain generate PIC code for
+      // additional security
+      // features (like ASLR). We default to PIC code to avoid linking issues on
+      // these OSes.
+      // On Android, PIC is default as well.
       relocModel = llvm::Reloc::PIC_;
     } else {
       // ARM for other than Darwin or Android defaults to static
@@ -571,6 +599,8 @@ llvm::TargetMachine *createTargetMachine(
   case FloatABI::Soft:
 #if LDC_LLVM_VER < 307
     targetOptions.UseSoftFloat = true;
+#else
+    features.AddFeature("+soft-float");
 #endif
     targetOptions.FloatABIType = llvm::FloatABI::Soft;
     break;
@@ -603,3 +633,13 @@ llvm::TargetMachine *createTargetMachine(
                                      targetOptions, relocModel, codeModel,
                                      codeGenOptLevel);
 }
+    
+ComputeBackend::Type getComputeTargetType(llvm::Module* m) {
+  llvm::Triple::ArchType a = llvm::Triple(m->getTargetTriple()).getArch();
+  if (a == llvm::Triple::spir || a == llvm::Triple::spir64)
+    return ComputeBackend::SPIRV;
+  else if (a == llvm::Triple::nvptx || a == llvm::Triple::nvptx64)
+    return ComputeBackend::NVPTX;
+  else
+    return ComputeBackend::None;
+}
