@@ -19,7 +19,9 @@
 #include "llvm/Support/Compiler.h"
 #include "gen/llvmcompat.h"
 
-#if LDC_LLVM_VER < 306
+#if LDC_LLVM_VER >= 500
+#define LLVM_END_WITH_NULL
+#elif LDC_LLVM_VER < 306
 #define LLVM_END_WITH_NULL END_WITH_NULL
 #endif
 
@@ -87,18 +89,28 @@ public:
 
   // Implement virtual functions needed by generic_parser_base
   unsigned getNumOptions() const LLVM_OVERRIDE { return 1; }
-  const char *getOption(unsigned N) const LLVM_OVERRIDE {
+#if LDC_LLVM_VER >= 400
+  llvm::StringRef
+#else
+  const char *
+#endif
+  getOption(unsigned N) const LLVM_OVERRIDE {
     assert(N == 0);
-#if LDC_LLVM_VER >= 308
+#if LDC_LLVM_VER >= 308 && LDC_LLVM_VER < 400
     return owner().ArgStr.data();
 #else
     return owner().ArgStr;
 #endif
   }
 
-  const char *getDescription(unsigned N) const LLVM_OVERRIDE {
+#if LDC_LLVM_VER >= 400
+  llvm::StringRef
+#else
+  const char *
+#endif
+  getDescription(unsigned N) const LLVM_OVERRIDE {
     assert(N == 0);
-#if LDC_LLVM_VER >= 308
+#if LDC_LLVM_VER >= 308 && LDC_LLVM_VER < 400
     return owner().HelpStr.data();
 #else
     return owner().HelpStr;
@@ -196,4 +208,16 @@ public:
 };
 }
 
+
+#if LDC_LLVM_VER >= 400
+#define clEnumValues llvm::cl::values
+#else
+template <typename DataType, typename... OptsTy>
+llvm::cl::ValuesClass<DataType> clEnumValues(const char *Arg, DataType Val,
+                                             const char *Desc,
+                                             OptsTy... Options) {
+  return llvm::cl::values(Arg, Val, Desc, Options..., clEnumValEnd);
+}
+#endif
+
 #endif
