@@ -39,6 +39,9 @@
 #include "llvm/Support/Debug.h"
 #include "llvm/Support/MathExtras.h"
 #include "llvm/Support/raw_ostream.h"
+#if LDC_LLVM_VER >= 500
+#include "llvm/Support/KnownBits.h"
+#endif
 
 #include <algorithm>
 
@@ -51,7 +54,8 @@ STATISTIC(NumDeleted,
           "Number of GC calls deleted because the return value was unused");
 
 static cl::opt<unsigned>
-    SizeLimit("dgc2stack-size-limit", cl::init(1024), cl::Hidden,
+    SizeLimit("dgc2stack-size-limit", cl::ZeroOrMore, cl::Hidden,
+              cl::init(1024),
               cl::desc("Require allocs to be smaller than n bytes to be "
                        "promoted, 0 to ignore."));
 
@@ -115,8 +119,15 @@ public:
   virtual Value *promote(CallSite CS, IRBuilder<> &B, const Analysis &A) {
     NumGcToStack++;
 
-    Instruction *Begin = &(*CS.getCaller()->getEntryBlock().begin());
-    return new AllocaInst(Ty, ".nongc_mem", Begin); // FIXME: align?
+    auto &BB = CS.getCaller()->getEntryBlock();
+    Instruction *Begin = &(*BB.begin());
+
+    // FIXME: set alignment on alloca?
+    return new AllocaInst(Ty,
+#if LDC_LLVM_VER >= 500
+                          BB.getModule()->getDataLayout().getAllocaAddrSpace(),
+#endif
+                          ".nongc_mem", Begin);
   }
 
   explicit FunctionInfo(ReturnType::Type returnType) : ReturnType(returnType) {}
@@ -138,16 +149,23 @@ static bool isKnownLessThan(Value *Val, uint64_t Limit, const Analysis &A) {
   if (Bits > BitsLimit) {
     APInt Mask = APInt::getLowBitsSet(Bits, BitsLimit);
     Mask.flipAllBits();
+#if LDC_LLVM_VER >= 500
+    KnownBits Known(Bits);
+    computeKnownBits(Val, Known, A.DL);
+    if ((Known.Zero & Mask) != Mask) {
+      return false;
+    }
+#else
     APInt KnownZero(Bits, 0), KnownOne(Bits, 0);
 #if LDC_LLVM_VER >= 307
     computeKnownBits(Val, KnownZero, KnownOne, A.DL);
 #else
     computeKnownBits(Val, KnownZero, KnownOne, &A.DL);
 #endif
-
     if ((KnownZero & Mask) != Mask) {
       return false;
     }
+#endif
   }
 
   return true;
@@ -255,7 +273,7 @@ public:
   }
 };
 
-// FunctionInfo for _d_newclass
+// FunctionInfo for _d_allocclass
 class AllocClassFI : public FunctionInfo {
 public:
   bool analyze(CallSite CS, const Analysis &A) override {
@@ -436,7 +454,7 @@ GarbageCollect2Stack::GarbageCollect2Stack()
   KnownFunctions["_d_allocmemoryT"] = &AllocMemoryT;
   KnownFunctions["_d_newarrayU"] = &NewArrayU;
   KnownFunctions["_d_newarrayT"] = &NewArrayT;
-  KnownFunctions["_d_newclass"] = &AllocClass;
+  KnownFunctions["_d_allocclass"] = &AllocClass;
   KnownFunctions["_d_allocmemory"] = &AllocMemory;
 }
 
@@ -474,7 +492,7 @@ bool GarbageCollect2Stack::runOnFunction(Function &F) {
   const DataLayout &DL = F.getParent()->getDataLayout();
   DominatorTree &DT = getAnalysis<DominatorTreeWrapperPass>().getDomTree();
   CallGraphWrapperPass *CGPass = getAnalysisIfAvailable<CallGraphWrapperPass>();
-  CallGraph *CG = CGPass ? &CGPass->getCallGraph() : 0;
+  CallGraph *CG = CGPass ? &CGPass->getCallGraph() : nullptr;
 #else
   DataLayoutPass *DLP = getAnalysisIfAvailable<DataLayoutPass>();
   assert(DLP && "required DataLayoutPass is null");
@@ -821,7 +839,8 @@ bool isSafeToStackAllocateArray(
 /// the attribute has to be removed before promoting the memory to the
 /// stack. The affected instructions are added to RemoveTailCallInsts. If
 /// the function returns false, these entries are meaningless.
-bool isSafeToStackAllocate(BasicBlock::iterator Alloc, Value *V, DominatorTree &DT,
+bool isSafeToStackAllocate(BasicBlock::iterator Alloc, Value *V,
+                           DominatorTree &DT,
                            SmallVector<CallInst *, 4> &RemoveTailCallInsts) {
   assert(isa<PointerType>(V->getType()) && "Allocated value is not a pointer?");
 
@@ -862,7 +881,13 @@ bool isSafeToStackAllocate(BasicBlock::iterator Alloc, Value *V, DominatorTree &
       CallSite::arg_iterator B = CS.arg_begin(), E = CS.arg_end();
       for (CallSite::arg_iterator A = B; A != E; ++A) {
         if (A->get() == V) {
-          if (!CS.paramHasAttr(A - B + 1, LLAttribute::NoCapture)) {
+#if LDC_LLVM_VER < 500
+          const unsigned paramHasAttr_firstArg = 1;
+#else
+          const unsigned paramHasAttr_firstArg = 0;
+#endif
+          if (!CS.paramHasAttr(A - B + paramHasAttr_firstArg,
+                               LLAttribute::NoCapture)) {
             // The parameter is not marked 'nocapture' - captured.
             return false;
           }
