@@ -15,6 +15,7 @@
 #include "gen/abi.h"
 #include "gen/classes.h"
 #include "gen/dvalue.h"
+#include "gen/funcgenstate.h"
 #include "gen/functions.h"
 #include "gen/irstate.h"
 #include "gen/llvm.h"
@@ -196,8 +197,7 @@ static void addExplicitArguments(std::vector<LLValue *> &args, AttrSet &attrs,
     }
 
     args[llArgIdx] = llVal;
-    // +1 as index 0 contains the function attributes.
-    attrs.add(llArgIdx + 1, irArg->attrs);
+    attrs.addToParam(llArgIdx, irArg->attrs);
 
     if (isVararg) {
       delete irArg;
@@ -207,18 +207,16 @@ static void addExplicitArguments(std::vector<LLValue *> &args, AttrSet &attrs,
 
 ////////////////////////////////////////////////////////////////////////////////
 
-static LLValue *getTypeinfoArrayArgumentForDVarArg(Expressions *arguments,
-                                                   int begin) {
+static LLValue *
+getTypeinfoArrayArgumentForDVarArg(const std::vector<DValue *> &argvals,
+                                   int begin) {
   IF_LOG Logger::println("doing d-style variadic arguments");
   LOG_SCOPE
 
   // number of non variadic args
   IF_LOG Logger::println("num non vararg params = %d", begin);
 
-  // get n args in arguments list
-  size_t n_arguments = arguments ? arguments->dim : 0;
-
-  const size_t numVariadicArgs = n_arguments - begin;
+  const size_t numVariadicArgs = argvals.size() - begin;
 
   // build type info array
   LLType *typeinfotype = DtoType(Type::dtypeinfo->type);
@@ -231,9 +229,9 @@ static LLValue *getTypeinfoArrayArgumentForDVarArg(Expressions *arguments,
   IF_LOG Logger::cout() << "_arguments storage: " << *typeinfomem << '\n';
 
   std::vector<LLConstant *> vtypeinfos;
-  vtypeinfos.reserve(n_arguments);
-  for (size_t i = begin; i < n_arguments; i++) {
-    vtypeinfos.push_back(DtoTypeInfoOf((*arguments)[i]->type));
+  vtypeinfos.reserve(argvals.size());
+  for (size_t i = begin; i < argvals.size(); i++) {
+    vtypeinfos.push_back(DtoTypeInfoOf(argvals[i]->type));
   }
 
   // apply initializer
@@ -328,11 +326,18 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
 
   // fence instruction
   if (fndecl->llvmInternal == LLVMfence) {
-    if (e->arguments->dim != 1) {
-      e->error("fence instruction expects 1 arguments");
+    if (e->arguments->dim < 1 || e->arguments->dim > 2) {
+      e->error("fence instruction expects 1 (or 2) arguments");
       fatal();
     }
-    p->ir->CreateFence(llvm::AtomicOrdering((*e->arguments)[0]->toInteger()));
+    auto atomicOrdering =
+        static_cast<llvm::AtomicOrdering>((*e->arguments)[0]->toInteger());
+    auto scope = llvm::SynchronizationScope::CrossThread;
+    if (e->arguments->dim == 2) {
+      scope = static_cast<llvm::SynchronizationScope>(
+          (*e->arguments)[1]->toInteger());
+    }
+    p->ir->CreateFence(atomicOrdering, scope);
     return true;
   }
 
@@ -596,7 +601,7 @@ bool DtoLowerMagicIntrinsic(IRState *p, FuncDeclaration *fndecl, CallExp *e,
 
     Expression *exp1 = (*e->arguments)[0];
     LLValue *ptr = DtoRVal(exp1);
-    result = new DImValue(exp1->type, DtoVolatileLoad(ptr));
+    result = new DImValue(e->type, DtoVolatileLoad(ptr));
     return true;
   }
 
@@ -626,13 +631,15 @@ public:
 
   ImplicitArgumentsBuilder(std::vector<LLValue *> &args, AttrSet &attrs,
                            Loc &loc, DValue *fnval,
-                           LLFunctionType *llCalleeType, Expressions *arguments,
-                           Type *resulttype, LLValue *retvar)
-      : args(args), attrs(attrs), loc(loc), fnval(fnval), arguments(arguments),
-        resulttype(resulttype), retvar(retvar),
+                           LLFunctionType *llCalleeType,
+                           const std::vector<DValue *> &argvals,
+                           Type *resulttype, LLValue *sretPointer)
+      : args(args), attrs(attrs), loc(loc), fnval(fnval), argvals(argvals),
+        resulttype(resulttype), sretPointer(sretPointer),
         // computed:
-        calleeType(fnval->type), dfnval(fnval->isFunc()),
-        irFty(DtoIrTypeFunction(fnval)), tf(DtoTypeFunction(fnval)),
+        isDelegateCall(fnval->type->toBasetype()->ty == Tdelegate),
+        dfnval(fnval->isFunc()), irFty(DtoIrTypeFunction(fnval)),
+        tf(DtoTypeFunction(fnval)),
         llArgTypesBegin(llCalleeType->param_begin()) {}
 
   void addImplicitArgs() {
@@ -653,12 +660,12 @@ private:
   AttrSet &attrs;
   Loc &loc;
   DValue *const fnval;
-  Expressions *const arguments;
+  const std::vector<DValue *> &argvals;
   Type *const resulttype;
-  LLValue *const retvar;
+  LLValue *const sretPointer;
 
   // computed:
-  Type *const calleeType;
+  const bool isDelegateCall;
   DFuncValue *const dfnval;
   IrFuncTy &irFty;
   TypeFunction *const tf;
@@ -673,17 +680,18 @@ private:
     size_t index = args.size();
     LLType *llArgType = *(llArgTypesBegin + index);
 
-    LLValue *var = retvar;
-    if (!var) {
-      var = DtoRawAlloca(llArgType->getContainedType(0),
-                         DtoAlignment(resulttype), ".rettmp");
+    LLValue *pointer = sretPointer;
+    if (!pointer) {
+      pointer = DtoRawAlloca(llArgType->getContainedType(0),
+                             DtoAlignment(resulttype), ".sret_tmp");
     }
 
-    args.push_back(var);
-    attrs.add(index + 1, irFty.arg_sret->attrs);
+    args.push_back(pointer);
+    attrs.addToParam(index, irFty.arg_sret->attrs);
 
     // verify that sret and/or inreg attributes are set
     const AttrBuilder &sretAttrs = irFty.arg_sret->attrs;
+    (void)sretAttrs;
     assert((sretAttrs.contains(LLAttribute::StructRet) ||
             sretAttrs.contains(LLAttribute::InReg)) &&
            "Sret arg not sret or inreg?");
@@ -692,10 +700,9 @@ private:
   // Adds an optional context/this pointer argument.
   void addContext() {
     bool thiscall = irFty.arg_this;
-    bool delegatecall = (calleeType->toBasetype()->ty == Tdelegate);
     bool nestedcall = irFty.arg_nest;
 
-    if (!thiscall && !delegatecall && !nestedcall)
+    if (!thiscall && !isDelegateCall && !nestedcall)
       return;
 
     size_t index = args.size();
@@ -726,7 +733,7 @@ private:
       // ... or a normal 'this' argument
       LLValue *thisarg = DtoBitCast(dfnval->vthis, llArgType);
       args.push_back(thisarg);
-    } else if (delegatecall) {
+    } else if (isDelegateCall) {
       // ... or a delegate context arg
       LLValue *ctxarg;
       if (fnval->isLVal()) {
@@ -752,14 +759,14 @@ private:
 
     // add attributes
     if (irFty.arg_this) {
-      attrs.add(index + 1, irFty.arg_this->attrs);
+      attrs.addToParam(index, irFty.arg_this->attrs);
     } else if (irFty.arg_nest) {
-      attrs.add(index + 1, irFty.arg_nest->attrs);
+      attrs.addToParam(index, irFty.arg_nest->attrs);
     }
 
     if (irFty.arg_objcSelector && dfnval) {
       if (auto sel = dfnval->func->objc.selector) {
-        LLGlobalVariable* selptr = objc_getMethVarRef(*sel);
+        LLGlobalVariable *selptr = objc_getMethVarRef(*sel);
         args.push_back(DtoBitCast(DtoLoad(selptr), getVoidPtrType()));
         hasObjcSelector = true;
       }
@@ -774,27 +781,26 @@ private:
 
     int numFormalParams = Parameter::dim(tf->parameters);
     LLValue *argumentsArg =
-        getTypeinfoArrayArgumentForDVarArg(arguments, numFormalParams);
+        getTypeinfoArrayArgumentForDVarArg(argvals, numFormalParams);
 
     args.push_back(argumentsArg);
-    attrs.add(args.size(), irFty.arg_arguments->attrs);
+    attrs.addToParam(args.size() - 1, irFty.arg_arguments->attrs);
   }
 };
 
 ////////////////////////////////////////////////////////////////////////////////
 
+namespace {
+
 // FIXME: this function is a mess !
-
-DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
-                        Expressions *arguments, llvm::Value *retvar) {
-  IF_LOG Logger::println("DtoCallFunction()");
-  LOG_SCOPE
-
+DValue *DtoCallFunctionImpl(Loc &loc, Type *resulttype, DValue *fnval,
+                            const std::vector<DValue *> &argvals,
+                            LLValue *sretPointer) {
   // make sure the D callee type has been processed
   DtoType(fnval->type);
 
   // get func value if any
-  DFuncValue *dfnval = fnval->isFunc();
+  DFuncValue *const dfnval = fnval->isFunc();
 
   // get function type info
   IrFuncTy &irFty = DtoIrTypeFunction(fnval);
@@ -820,14 +826,14 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   AttrSet attrs;
 
   // return attrs
-  attrs.add(0, irFty.ret->attrs);
+  attrs.addToReturn(irFty.ret->attrs);
 
   std::vector<LLValue *> args;
   args.reserve(irFty.args.size());
 
   // handle implicit arguments (sret, context/this, _arguments)
-  ImplicitArgumentsBuilder iab(args, attrs, loc, fnval, callableTy, arguments,
-                               resulttype, retvar);
+  ImplicitArgumentsBuilder iab(args, attrs, loc, fnval, callableTy, argvals,
+                               resulttype, sretPointer);
   iab.addImplicitArgs();
 
   // handle explicit arguments
@@ -845,34 +851,6 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   }
 
   const int numFormalParams = Parameter::dim(tf->parameters); // excl. variadics
-  const size_t n_arguments =
-      arguments ? arguments->dim : 0; // number of explicit arguments
-
-  std::vector<DValue *> argvals(n_arguments, static_cast<DValue *>(nullptr));
-  if (dfnval && dfnval->func->isArrayOp) {
-    // For array ops, the druntime implementation signatures are crafted
-    // specifically such that the evaluation order is as expected with
-    // the strange DMD reverse parameter passing order. Thus, we need
-    // to actually build the arguments right-to-left for them.
-    for (int i = numFormalParams - 1; i >= 0; --i) {
-      Parameter *fnarg = Parameter::getNth(tf->parameters, i);
-      assert(fnarg);
-      DValue *argval = DtoArgument(fnarg, (*arguments)[i]);
-      argvals[i] = argval;
-    }
-  } else {
-    for (int i = 0; i < numFormalParams; ++i) {
-      Parameter *fnarg = Parameter::getNth(tf->parameters, i);
-      assert(fnarg);
-      DValue *argval = DtoArgument(fnarg, (*arguments)[i]);
-      argvals[i] = argval;
-    }
-  }
-  // add varargs
-  for (size_t i = numFormalParams; i < n_arguments; ++i) {
-    argvals[i] = DtoArgument(nullptr, (*arguments)[i]);
-  }
-
   addExplicitArguments(args, attrs, irFty, callableTy, argvals,
                        numFormalParams);
 
@@ -885,7 +863,17 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   }
 
   // call the function
-  LLCallSite call = gIR->func()->scopes->callOrInvoke(callable, args);
+  LLCallSite call =
+      gIR->funcGen().callOrInvoke(callable, args, "", tf->isnothrow);
+
+#if LDC_LLVM_VER >= 309
+  // PGO: Insert instrumentation or attach profile metadata at indirect call
+  // sites.
+  if (!call.getCalledFunction()) {
+    auto &PGO = gIR->funcGen().pgo;
+    PGO.emitIndirectCallPGO(call.getInstruction(), callable);
+  }
+#endif
 
   // get return value
   const int sretArgIndex =
@@ -893,7 +881,8 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
                                                                         : 0);
   LLValue *retllval =
       (irFty.arg_sret ? args[sretArgIndex] : call.getInstruction());
-  bool retValIsLVal = (tf->isref && returnTy != Tvoid) || (irFty.arg_sret != nullptr);
+  bool retValIsLVal =
+      (tf->isref && returnTy != Tvoid) || (irFty.arg_sret != nullptr);
 
   if (!retValIsLVal) {
     // let the ABI transform the return value back
@@ -990,7 +979,7 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
   }
 
   // set calling convention and parameter attributes
-  llvm::AttributeSet &attrlist = attrs;
+  LLAttributeSet &attrlist = attrs;
   if (dfnval && dfnval->func) {
     LLFunction *llfunc = llvm::dyn_cast<LLFunction>(DtoRVal(dfnval));
     if (llfunc && llfunc->isIntrinsic()) // override intrinsic attrs
@@ -1005,9 +994,14 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
     call.setCallingConv(callconv);
   }
   // merge in function attributes set in callOrInvoke
+#if LDC_LLVM_VER >= 500
   attrlist = attrlist.addAttributes(
-      gIR->context(), llvm::AttributeSet::FunctionIndex, call.getAttributes());
-
+      gIR->context(), LLAttributeSet::FunctionIndex,
+      llvm::AttrBuilder(call.getAttributes(), LLAttributeSet::FunctionIndex));
+#else
+  attrlist = attrlist.addAttributes(
+      gIR->context(), LLAttributeSet::FunctionIndex, call.getAttributes());
+#endif
   call.setAttributes(attrlist);
 
   // Special case for struct constructor calls: For temporaries, using the
@@ -1026,3 +1020,57 @@ DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
 
   return new DImValue(resulttype, retllval);
 }
+
+std::vector<DValue *> evaluateArgExpressions(DValue *fnval,
+                                             Expressions *arguments) {
+  IF_LOG Logger::println("Evaluating argument expressions");
+  LOG_SCOPE
+
+  const auto tf = DtoTypeFunction(fnval);
+
+  const size_t numArguments = arguments ? arguments->dim : 0;
+  std::vector<DValue *> argvals(numArguments);
+
+  // For array ops, the druntime implementation signatures are crafted
+  // specifically such that the evaluation order is as expected with
+  // the strange DMD reverse parameter passing order. Thus, we need
+  // to actually evaluate the arguments right-to-left for them.
+  const auto dfnval = fnval->isFunc();
+  const bool reverse = (dfnval && dfnval->func && dfnval->func->isArrayOp);
+
+  // formal params (excl. variadics)
+  const size_t numFormalParams = Parameter::dim(tf->parameters);
+  for (size_t j = 0; j < numFormalParams; ++j) {
+    size_t i = (!reverse ? j : numFormalParams - 1 - j);
+    Parameter *fnarg = Parameter::getNth(tf->parameters, i);
+    assert(fnarg);
+    argvals[i] = DtoArgument(fnarg, (*arguments)[i]);
+  }
+
+  // append variadics
+  for (size_t i = numFormalParams; i < numArguments; ++i) {
+    argvals[i] = DtoArgument(nullptr, (*arguments)[i]);
+  }
+
+  return argvals;
+}
+
+} // anonymous namespace
+
+DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
+                        const std::vector<DValue *> &argvals,
+                        LLValue *sretPointer) {
+  IF_LOG Logger::println("DtoCallFunction()");
+  LOG_SCOPE
+
+  return DtoCallFunctionImpl(loc, resulttype, fnval, argvals, sretPointer);
+}
+
+DValue *DtoCallFunction(Loc &loc, Type *resulttype, DValue *fnval,
+                        Expressions *arguments, llvm::Value *sretPointer) {
+  IF_LOG Logger::println("DtoCallFunction()");
+  LOG_SCOPE
+
+  const auto argvals = evaluateArgExpressions(fnval, arguments);
+  return DtoCallFunctionImpl(loc, resulttype, fnval, argvals, sretPointer);
+}
